package nu.dll.duper;

import java.util.*;
import java.io.*;
import java.security.*;
import java.nio.*;
import java.nio.channels.*;

public class Duper {

    Set beenThere = new HashSet();
    Map sizeFileMap = new HashMap();
    Map fileChecksumMap = new HashMap();
    List dupes = new LinkedList();
    List roots;
    MessageDigest md;
    long minimumSize = 32;
    int filesProcessed = 0;
    int filesSkipped = 0;
    int directoriesProcessed = 0;
    boolean pedantic = false;

    boolean aborted = false;

    ProgressListener listener = null;

    long startTime, endTime;

    int dupeCount = 0;

    long totalBytes = 0;

    long totalBytesChecksummed = 0;
    int totalFilesChecksummed = 0;
    Map checksumFileMap = new HashMap();

    int blocksize = 5120;
    byte[] md5buffer = new byte[blocksize];

    List folderIgnorePatterns = new LinkedList();
    List fileIgnorePatterns = new LinkedList();

    boolean ignoreHidden = true;

    public interface ProgressListener {
	public void scanStart();
	public void scanStop(boolean aborted);
	public void enteringDirectory(File f);
    }

    public void setProgressListener(ProgressListener listener) {
	this.listener = listener;
    }

    public void addFileIgnorePattern(String s) {
	fileIgnorePatterns.add(s);
    }

    public void addFolderIgnorePattern(String s) {
	folderIgnorePatterns.add(s);
    }

    public Duper(List roots) {
	this.roots = roots;
	try {
	    md = MessageDigest.getInstance("MD5");
	} catch (NoSuchAlgorithmException ex1) {
	    throw new RuntimeException(ex1);
	}
    }

    public void processRoots() throws IOException {
	startTime = System.currentTimeMillis();
	endTime = 0;
	if (listener != null) listener.scanStart();
	for (Iterator i = roots.iterator();!aborted && i.hasNext();) {
	    processDirectory((File) i.next());
	}
	sparehack();
	if (!aborted && pedantic) {
	    pedantic();
	}
	endTime = System.currentTimeMillis();
	if (listener != null) listener.scanStop(aborted);
    }

    private void sparehack() {
	for (Iterator i = dupes.iterator();i.hasNext();) {
	    List l = (List) i.next();
	    if (l.size() == 1) {
		i.remove();
	    } else {
		dupeCount += l.size()-1;
	    }
	    
	}
    }

    public long getDuration() {
	if (endTime == 0) {
	    return System.currentTimeMillis() - startTime;
	} else {
	    return endTime - startTime;
	}
    }

    void pedantic() throws IOException {
	for (Iterator i = dupes.iterator();i.hasNext();) {
	    List files = (List) i.next();
	    if (files.size() == 2) {
		if (!pedanticCompareFiles((File) files.get(0), (File) files.get(1))) {
		    i.remove();
		}
	    } else {
		Object[] filesArr = files.toArray();
		List toRemove = new LinkedList();
		for (int j=0; j < filesArr.length; j++) {
		    for (int k=j+1; k < filesArr.length; k++) {
			if (!pedanticCompareFiles((File) filesArr[j], (File) filesArr[k])) {
			    toRemove.add(filesArr[k]); // XXX help, what to do? when/which to remove?
			}
		    }
		}
		files.removeAll(toRemove);
		throw new RuntimeException("Size: " + files.size() + ": Not supported yet.");
	    }
	}
    }

    boolean pedanticCompareFiles(File a, File b) throws IOException {
	dprintln("pedantic: " + a + " vs " + b);
	FileInputStream stream_a = new FileInputStream(a);
	FileInputStream stream_b = new FileInputStream(b);
	FileChannel chan_a = stream_a.getChannel();
	FileChannel chan_b = stream_b.getChannel();
	ByteBuffer buf_a = chan_a.map(FileChannel.MapMode.READ_ONLY, 0, chan_a.size());
	ByteBuffer buf_b = chan_b.map(FileChannel.MapMode.READ_ONLY, 0, chan_b.size());
	boolean equal = true;
	while (equal && buf_a.hasRemaining() && buf_b.hasRemaining()) {
	    if (buf_a.get() != buf_b.get()) {
		equal = false;
	    }
	}
	if (buf_a.hasRemaining() || buf_b.hasRemaining()) {
	    equal = false;
	}
	dprintln("pedantic check failed");
	chan_a.close();
	chan_b.close();
	return equal;
    }

    boolean isRoot(File file) {
	File[] roots = File.listRoots();
	for (int i=0; i < roots.length; i++) {
	    if (file.equals(roots[i])) return true;
	}
	return false;
    }

    void processDirectory(File dir) throws IOException {
	if (!dir.isDirectory())
	    throw new IOException(dir.getAbsolutePath() + " isn't a directory");
	if (ignoreHidden && dir.isHidden() && !isRoot(dir)) {
	    dprintln("Skipping hidden directory " + dir);
	    return;
	}

	if (listener != null) {
	    listener.enteringDirectory(dir);
	}
	File[] files = dir.listFiles();

	for (int i=0; !aborted && files != null && i < files.length; i++) {
	    if (files[i].isDirectory()) {
		processDirectory(files[i]);
	    } else {
		processFile(files[i]);
	    }
	}
	directoriesProcessed++;
    }

    void abort() {
	aborted = true;
    }


    void processFile(File file) throws IOException {
	Long size = new Long(file.length());
	if (size.longValue() < minimumSize) {
	    filesSkipped++;
	    return;
	}
	totalBytes += size.longValue();
	if (!sizeFileMap.containsKey(size)) {
	    List files = new LinkedList();
	    files.add(file);
	    sizeFileMap.put(size, files);
	    filesProcessed++;
	    return;
	}
	List files = (List) sizeFileMap.get(size);
	files.add(file);
	//dprintln("Size match (" + size + "): " + files);
	compareFiles(files);
	filesProcessed++;
    }

    boolean equals(byte[] a, byte[] b) {
	return MessageDigest.isEqual(a, b);
    }

    void dprintln(String s) {
	System.out.println(s);
    }

    void compareFiles(List files) throws IOException {
	for (Iterator i = files.iterator(); i.hasNext();) {
	    File f = (File) i.next();
	    String md5string = getMD5(f);
	    //dprintln("MD5 for " + f + ": " + md5string);

	    List sumfiles = (List) checksumFileMap.get(md5string);
	    if (sumfiles == null) {
		sumfiles = new LinkedList();
		checksumFileMap.put(md5string, sumfiles);
	    } else {
		if (!dupes.contains(sumfiles)) dupes.add(sumfiles);
		//dprintln("(dupe!)");
	    }
	    if (!sumfiles.contains(f)) sumfiles.add(f);
	}
    }

    String hexstring(byte[] data) {
	StringBuffer buf = new StringBuffer();
	for (int i=0; i < data.length; i++) {
	    if (((int)  data[i] & 0xff) < 0x10) buf.append("0");
	    buf.append(Long.toString((int) data[i] & 0xff, 16));
	}
	return buf.toString();
    }

	
    byte[] calcMD5(ByteBuffer buf) throws IOException {
	md.reset();
	while (buf.hasRemaining()) {
	    int remaining = buf.remaining();
	    if (remaining >= blocksize) {
		buf.get(md5buffer);
		md.update(md5buffer);
		totalBytesChecksummed += blocksize;
	    } else {
		buf.get(md5buffer, 0, remaining);
		md.update(md5buffer, 0, remaining);
		totalBytesChecksummed += remaining;
	    }
	}
	totalFilesChecksummed++;
	return md.digest();
    }

    public String getMD5(File file) throws IOException {
	if (fileChecksumMap.containsKey(file)) {
	    return (String) fileChecksumMap.get(file);
	}
	FileInputStream fis = new FileInputStream(file);
	FileChannel fc = fis.getChannel();
	int size = (int) fc.size();
	MappedByteBuffer buf = fc.map(FileChannel.MapMode.READ_ONLY, 0, size);
	byte[] sum = calcMD5(buf);
	fc.close();
	fis.close();
	String sumstr = hexstring(sum);
	fileChecksumMap.put(file, sumstr);
	return sumstr;
    }

    public List getDupes() {
	return dupes;
    }

    public static void main(String[] argv) throws Exception {
	List roots = new LinkedList();
	roots.add(new File("E:\\Musik"));
	Duper d = new Duper(roots);
	d.processRoots();
	System.out.println("******");
	for (Iterator i = d.dupes.iterator();i.hasNext();) {
	    List dupes = (List) i.next();
	    File first = (File) dupes.get(0);
	    System.out.println("* " + first.getName());
	    for (Iterator ii = dupes.iterator();ii.hasNext();) {
		System.out.println("\t" + ((File) ii.next()).getAbsolutePath());
	    }
	}
    }


}
