package nu.dll.duper;

import org.eclipse.swt.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.browser.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.program.Program;

import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

import java.io.File;
import java.io.IOException;

import java.util.LinkedList;
import java.util.Iterator;

import java.util.Map;
import java.util.HashMap;

public class GUI implements Duper.ProgressListener {
    Display display;
    Shell shell;
    List rootList;
    List consoleList;
    Table dupeTable;

    Button startScanBtn;
    Button stopScanBtn;

    Duper duper = null;
    File homeDir = new File(System.getProperty("user.home"));
    File rootListFile = new File(homeDir, ".duper.roots");

    Exception exception;

    Color white;
    Color pyjamasColor;

    boolean isWindows = System.getProperty("os.name").startsWith("Windows");

    public GUI(Display display) {
	this.display = display;
	pyjamasColor = new Color(display, 200, 200, 200);
	white = new Color(display, 255, 255, 255);
    }

    void defaultRootList() {
	if (rootListFile.exists()) {
	    try {
		BufferedReader reader = 
		    new BufferedReader(new InputStreamReader(new FileInputStream(rootListFile), "utf-8"));
		String row;
		while ((row = reader.readLine()) != null) {
		    row = row.trim();
		    rootList.add(row);
		}
		reader.close();
	    } catch (IOException ex1) {
		throw new RuntimeException("I/O error: " + ex1.toString());
	    }
	} else {
	    
	    File[] roots = File.listRoots();
	    for (int i=0; i < roots.length; i++) {
		if (isWindows && roots[i].getAbsolutePath().startsWith("A:")) continue;
		rootList.add(roots[i].getAbsolutePath());
	    }
	}

    }

    synchronized void saveRootList() {
	try {
	    Writer writer = new OutputStreamWriter(new FileOutputStream(rootListFile), "utf-8");
	    String[] items = rootList.getItems();
	    for (int i=0; i < items.length; i++) {
		writer.write(items[i] + "\n");
	    }
	    writer.close();
	} catch (IOException ex1) {
	    throw new RuntimeException(ex1.toString());
	}
    }

    public void run() {
	shell = new Shell(display, SWT.CLOSE|SWT.MIN|SWT.MAX|SWT.BORDER|SWT.TITLE|SWT.RESIZE);
	GridLayout gl = new GridLayout();
	gl.numColumns = 2;
	shell.setLayout(gl);
	rootList = new List(shell, SWT.MULTI | SWT.V_SCROLL | SWT.BORDER);
	GridData gd = new GridData(GridData.FILL_HORIZONTAL);
	gd.horizontalSpan = 2;
	gd.heightHint = 100;
	rootList.setLayoutData(gd);

	defaultRootList();

	Button deleteRootBtn = new Button(shell, SWT.NONE);
	deleteRootBtn.setText("Remove folder(s)");
	deleteRootBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    rootList.remove(rootList.getSelectionIndices());
		    saveRootList();
		}
	    });

	Button addRootBtn = new Button(shell, SWT.NONE);
	addRootBtn.setText("Add folder");
	addRootBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    DirectoryDialog dialog = new DirectoryDialog(shell);
		    dialog.setMessage("Please choose a folder to add");
		    String selected = dialog.open();
		    if (selected != null) {
			rootList.add(selected);
			saveRootList();
		    }
		}
	    });

	startScanBtn = new Button(shell, SWT.NONE);
	startScanBtn.setText("Start scan!");
	startScanBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    startScanThread();
		}
	    });

	stopScanBtn = new Button(shell, SWT.NONE);
	stopScanBtn.setEnabled(false);
	stopScanBtn.setText("Stop scan");
	stopScanBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    stopScanThread();
		}
	    });
	
	
	consoleList = new List(shell, SWT.MULTI | SWT.V_SCROLL | SWT.H_SCROLL | SWT.BORDER);
	gd = new GridData(GridData.FILL_HORIZONTAL);
	gd.horizontalSpan = 2;
	gd.heightHint = 100;
	gd.widthHint = 400;
	consoleList.setLayoutData(gd);

	dupeTable = new Table(shell, SWT.CHECK | SWT.V_SCROLL | SWT.H_SCROLL | SWT.BORDER |
			      SWT.FULL_SELECTION);
	dupeTable.setHeaderVisible(true);

	gd = new GridData(GridData.FILL_BOTH);
	gd.horizontalSpan = 2;
	gd.heightHint = 250;
	dupeTable.setLayoutData(gd);


	TableColumn md5col = new TableColumn(dupeTable, SWT.LEFT);
	md5col.setText("MD5");
	md5col.setWidth(220);
	
	TableColumn namecol = new TableColumn(dupeTable, SWT.LEFT);
	namecol.setText("Filename");
	namecol.setWidth(300);

	TableColumn sizecol = new TableColumn(dupeTable, SWT.LEFT);
	sizecol.setText("Size");
	sizecol.setWidth(50);

	Composite composite = new Composite(shell, SWT.NONE);
	gd = new GridData(GridData.FILL_HORIZONTAL);
	gd.horizontalSpan = 2;
	composite.setLayoutData(gd);
	composite.setLayout(new FillLayout());

	Button uncheckAllBtn = new Button(composite, SWT.NONE);
	uncheckAllBtn.setText("Uncheck all files");
	uncheckAllBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    checkAll(false);
		}
	    });


	Button checkAllBtn = new Button(composite, SWT.NONE);
	checkAllBtn.setText("Check all files");
	checkAllBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    checkAll(true);
		}
	    });

	Button checkAllButFirstBtn = new Button(composite, SWT.NONE);
	checkAllButFirstBtn.setText("Check all but first");
	checkAllButFirstBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    checkAllButFirst();
		}
	    });
	

	Button checkAllButLastBtn = new Button(composite, SWT.NONE);
	checkAllButLastBtn.setText("Check all but last");
	checkAllButLastBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    checkAllButLast();
		}
	    });

	Button deleteCheckedFilesBtn = new Button(shell, SWT.NONE);
	deleteCheckedFilesBtn.setText("Delete checked files!");
	deleteCheckedFilesBtn.addSelectionListener(new SelectionAdapter() {
		public void widgetSelected(SelectionEvent e) {
		    deleteCheckedFiles();
		}

	});
	gd = new GridData();
	gd.horizontalSpan = 2;
	deleteCheckedFilesBtn.setLayoutData(gd);

	shell.setText("Duper");
	shell.pack();
	shell.setVisible(true);
	while (!shell.isDisposed()) {
	    if (!display.readAndDispatch()) {
		display.sleep();
	    }
	}
	display.dispose();

    }

    void deleteCheckedFiles() {
	final LinkedList list = new LinkedList();
	final Map tableIndexMap = new HashMap();
	TableItem[] items = dupeTable.getItems();
	for (int i=0; i < items.length; i++) {
	    if (items[i].getChecked()) {
		File f = new File(items[i].getText(1));
		list.add(f);
		tableIndexMap.put(f, new Integer(i));
	    }
	}
	long totalSize = 0;
	for (Iterator i = list.iterator();i.hasNext();) {
	    totalSize += ((File) i.next()).length();
	}
	double megs = ((double)totalSize) / 1024d / 1024d;
	
	MessageBox mb = new MessageBox(shell, SWT.ICON_WARNING | SWT.YES | SWT.NO);
	mb.setText("Delete selected files");
	mb.setMessage("You have selected " + list.size() + " files, using " +
		      "a total of " + (megs > 1 ? (((int) megs) + " megabytes") : (totalSize + " bytes"))
		      + ". Are you SURE you want to delete those files?");
	if (mb.open() == SWT.YES) {
	    Thread t = new Thread() {
		    public void run() {
			reallyDeleteFiles(list, tableIndexMap);
		    }
		};
	    t.setDaemon(true);
	    t.start();
	}
    }

    void reallyDeleteFiles(LinkedList files, Map tableIndices) {
	boolean aborted = false;
	log("Deleting " + files.size() + " files...");
	int deleted = 0;
	long bytes = 0;
	LinkedList indicesDeleted = new LinkedList();
	for (Iterator i = files.iterator();!aborted && i.hasNext();) {
	    File f = (File) i.next();
	    long size = f.length();
	    if (f.delete()) {
		deleted++;
		bytes += size;
		indicesDeleted.add(tableIndices.get(f));
		log("Deleted " + f.getAbsolutePath());
	    } else {
		log("ERROR: Failed to delete " + f.getAbsolutePath());
	    }
	}
	final int indices[] = new int[indicesDeleted.size()];
	int c=0;
	for (Iterator i = indicesDeleted.iterator();i.hasNext();c++) {
	    indices[c] = ((Integer) i.next()).intValue();
	}
	display.asyncExec(new Runnable() {
		public void run() {
		    dupeTable.remove(indices);
		}
	    });

	log("Done: deleted " + deleted + " files (" + bytes + " bytes freed).");
    }

    void checkAllButLast() {
	checkAll(true);
	TableItem[] items = dupeTable.getItems();
	String lastMD5 = null;
	TableItem lastItem = null;
	for (int i=0; i < items.length; i++) {
	    String md5 = items[i].getText(0);
	    if (!md5.equals(lastMD5)) {
		if (lastItem != null) {
		    lastItem.setChecked(false);
		}
	    }
	    lastMD5 = md5;
	    lastItem = items[i];
	}
	if (items[items.length-2].getText(0).equals(items[items.length-1].getText(0))) {
	    items[items.length-1].setChecked(false);
	}
    }

    void checkAllButFirst() {
	TableItem[] items = dupeTable.getItems();
	String lastMD5 = null;
	for (int i=0; i < items.length; i++) {
	    String md5 = items[i].getText(0);
	    if (md5.equals(lastMD5)) {
		items[i].setChecked(true);
	    } else {
		items[i].setChecked(false);
	    }
	    lastMD5 = md5;
	}
    }

    void checkAll(boolean checked) {
	TableItem[] items = dupeTable.getItems();
	for (int i=0; i < items.length; i++) {
	    items[i].setChecked(checked);
	}
    }


    void startScanThread() {
	startScanBtn.setEnabled(false);
	stopScanBtn.setEnabled(true);
	LinkedList roots = new LinkedList();
	String[] items = rootList.getItems();
	for (int i=0; i < items.length; i++) {
	    roots.add(new File(items[i]));
	}
	duper = new Duper(roots);
	duper.setProgressListener(this);
	Thread t = new Thread() {
		public void run() {
		    try {
			duper.processRoots();
		    } catch (Exception ex1) {
			ex1.printStackTrace();
			exception = ex1;
			scanStop(true);
		    }
		}
	    };
	t.setDaemon(true);
	t.setName("Scan-Thread");
	t.start();
    }

    void log(final String s) {
	display.asyncExec(new Runnable() {
		public void run() {
		    consoleList.add(s);
		    consoleList.setSelection(consoleList.getItemCount()-1);
		    consoleList.showSelection();
		}
	    });
    }

    void showScanResults() {
	log("--------------------------------------------------------------------------");
	log("Files: " + duper.filesProcessed + ", directories: " + 
	    duper.directoriesProcessed + ", bytes: " + duper.totalBytes);
	log(duper.dupeCount + " duplicates found.");
	log("MD5 calculated for " + duper.totalFilesChecksummed + 
	    " files (" + duper.totalBytesChecksummed + " bytes)");
	long millis = duper.getDuration();
	long seconds = millis / 1000l;
	log("Duration: " + seconds + " seconds");
	log("--------------------------------------------------------------------------");

	dupeTable.removeAll();
	java.util.List dupelist = duper.getDupes();
	String lastmd5 = null;
	boolean pyjamas = false;
	for (Iterator i = dupelist.iterator();i.hasNext();) {
	    java.util.List files = (java.util.List) i.next();
	    for (Iterator j = files.iterator();j.hasNext();) {
		File f = (File) j.next();
		TableItem item = new TableItem(dupeTable, SWT.NONE);

		String name = f.getName();
		int lastDot = name.lastIndexOf(".");
		// if the file has an extension, check to see if it
		// has an icon associated with it.
		if (lastDot > 0 && lastDot != name.length()-1) {
		    String extension = name.substring(lastDot+1);
		    Program program = Program.findProgram(extension);
		    if (program != null) {
			ImageData id = program.getImageData();
			if (id != null) {
			    final Image image = new Image(display, id);
			    item.addDisposeListener(new DisposeListener() {
				    public void widgetDisposed(DisposeEvent e) {
					image.dispose();
				    }
				});
			    item.setImage(0, image);
			}
		    }
		}
		try {
		    String md5 = duper.getMD5(f);
		    if (!md5.equals(lastmd5)) {
			pyjamas = !pyjamas;
		    }
		    item.setText(0, md5);
		    lastmd5 = md5;
		} catch (IOException ex1) {
		    throw new RuntimeException("I/O error fetching MD5: " + ex1.toString());
		}
		item.setText(1, f.getAbsolutePath());
		item.setText(2, "" + f.length());
		item.setData(f);
		item.setChecked(false);
		if (pyjamas) {
		    item.setBackground(pyjamasColor);
		} else {
		    item.setBackground(white);
		}
	    }
	}
    }

    public void enteringDirectory(File f) {
	log("Scanning: " + f.getAbsolutePath());
    }

    public void scanStart() {
	log("Scan started.");
    }

    public void scanStop(boolean aborted) {
	if (aborted) {
	    if (exception != null) {
		log("Scan error: " + exception.toString());
		log("Scan was aborted due to an error.");
	    } else {
		log("Scan aborted by user.");
	    }
	} else {
	    log("Scan complete.");
	    display.asyncExec(new Runnable() {
		    public void run() {
			showScanResults();
		    }
		});
	}
	display.asyncExec(new Runnable() {
		public void run() {
		    startScanBtn.setEnabled(true);
		    stopScanBtn.setEnabled(false);
		}
	    });
    }

    void stopScanThread() {
	duper.abort();
    }
    
    public static void main(String[] argv) {
	new GUI(new Display()).run();
    }
}
